import time
from twisted.internet import defer


class Job(object):
    """
    Hold the details of a job that is being, or will be, processed by a
    L{ResizableDispatchQueue}.

    All jobs start out in the PENDING state. Then either of the following
    will eventually happen:

      1. The job is launched. State is set to UNDERWAY.

      2. The job is cancelled (self.cancel is called) without being
         launched.  The state is then set to CANCELLED.

    Once a job is UNDERWAY, the following may happen:

      1. The function that launches the job hits an error of some
         kind. State goes to FAILED.

      2. The job completes cleanly. State goes to FINISHED.

      3. The job (or, equivalently, the deferred that will receive the
         result) is cancelled. State is set to CANCELLED.

    In rough ASCII art the possible job state transitions are:

                               ->  FAILED
                              /
      PENDING  -->  UNDERWAY  -->  FINISHED
         |                    \
         ----------------------->  CANCELLED


    @ivar jobarg: The argument that will be passed to the function that
    processes the job.

    @ivar priority: The priority for the job. Lower means more important.

    @ivar queuedTime: The time the job entered the queue.

    @ivar startTime: The time processing on the job starts.

    @ivar stopTime: The time job processing stops. Processing stops due to
    one of: clean completion (state = FINISHED), failure during the
    processing (state = FAILED), or job cancellation (state = CANCELLED).

    @ivar finishedDeferred: a C{Deferred} that will fire when the job
    finished, with the job function result.

    @ivar state: The current state of the job, as described above.
    """

    PENDING = 0
    UNDERWAY = 1
    FINISHED = 2
    FAILED = 3
    CANCELLED = 4

    def __init__(self, jobarg, priority):
        self.jobarg = jobarg
        self.priority = priority
        self.queuedTime = time.time()
        self.startTime = None
        self.stopTime = None
        self.state = self.PENDING
        self.waiting = []  # Deferreds waiting on the result of this job.

    def launch(self, func):
        """
        Begin operation on this job by passing its job argument to C{func}.
        The call to func is made via defer.maybeDeferred.  When the job
        finishes, its result will be passed on to self.finishedDeferred in
        self._finish below. Arrange to cancel the in-progress work if
        cancel is called on the C{Deferred} we return.

        @param func: the function to call on self.jobarg.

        @raise: C{RuntimeError} if the job is not in the PENDING state.

        @return: A C{Deferred} that will fire (with self) after running
        func on self.jobarg. The result of that call will be in
        self.result.
        """
        if self.state != self.PENDING:
            raise RuntimeError('You cannot launch a job that is not pending.')
        else:
            self.startTime = time.time()
            self.state = self.UNDERWAY
            self._underwayDeferred = defer.maybeDeferred(func, self.jobarg)
            self._underwayDeferred.addCallbacks(self._finished, self._failed)

    def _finished(self, result):
        """
        The job finished cleanly. Record its finishing time, change state,
        set self.result, and pass on the result (as self, a completed job).
        """
        assert self.state == self.UNDERWAY
        self.stopTime = time.time()
        self.state = self.FINISHED
        self.result = result
        for d in self.waiting:
            d.callback(self)
        self.waiting = []

    def _failed(self, failure):
        """
        The job failed in some way (including possibly being cancelled). If
        it wasn't cancelled, record its finishing time and errback all
        waiting watchers.  If the job was cancelled, its completion details
        have already been set (by self.cancel, below).
        """
        if self.state != self.CANCELLED:
            assert self.state == self.UNDERWAY
            self.stopTime = time.time()
            self.state = self.FAILED
            self.failure = failure
            for d in self.waiting:
                d.errback(self)
            self.waiting = []

    def cancel(self):
        """
        Cancel the job, record the cancellation time, cancel the underway
        deferred (if any). Errback all waiting watchers.

        @param deferred: The deferred that is being cancelled (i.e.,
        self._underwayDeferred). Ignored.
        """
        self.stopTime = time.time()
        if self.state == self.UNDERWAY:
            self.state = self.CANCELLED
            self._underwayDeferred.cancel()
        self.state = self.CANCELLED
        for d in self.waiting:
            d.errback(self)
        self.waiting = []

    def _cancelWatcher(self, ignoredDeferred):
        """
        A C{Deferred} returned by self.watch has been cancelled.  Cancel
        the job by calling the public self.cancel method.

        @param ignoredDeferred: The deferred whose cancel method was
        called. Ignored.
        """
        self.cancel()

    def watch(self):
        """
        Watch this job.

        @return: a C{Deferred} that fires with the job result or fails if
        it is cancelled or hits an error.
        """
        if self.state in (self.PENDING, self.UNDERWAY):
            d = defer.Deferred(self._cancelWatcher)
            self.waiting.append(d)
            return d
        elif self.state == self.FINISHED:
            return defer.succeed(self)
        else:
            return defer.fail(self)



import itertools
from heapq import heappush, heappop
from collections import deque
from twisted.internet import defer


class _Task(object):
    """
    Holds details of a task in a C{DeferredPriorityQueue}.

    @ivar priority: The priority of the task. Lower is more important.

    @ivar count: Used to maintain a stable (sort) order of tasks that have
    the same priority.

    @ivar valid: if C{False}, this task is no longer valid (i.e., it has
    been deleted).

    @ivar obj: The object that was put into the C{DeferredPriorityQueue}.
    """
    def __init__(self, obj, priority, count):
        self.obj = obj
        self.priority = priority
        self.count = count
        self.valid = True

    def __lt__(self, other):
        return (self.priority, self.count) < (other.priority, other.count)


class DeferredPriorityQueue(object):
    """
    A prioritized event-driven queue.

    Objects may be added as usual to this queue.  When an attempt is made
    to retrieve an object when the queue is empty, a C{Deferred} is
    returned which will fire when an object becomes available.

    Note that the same object cannot currently be in the queue twice at the
    same time.  This is because the simple support for delete and
    reprioritize makes doing duplicated objects awkward. Attempts to re-put
    an object already in the queue will raise (see docs for put below).

    Implementation adapted from http://docs.python.org/library/heapq.html
    and from Twisted's twisted.internet.defer.DeferredQueue class.

    @ivar size: The maximum number of objects to allow into the queue at a
    time.  When an attempt to add a new object would exceed this limit,
    C{defer.QueueOverflow} is raised synchronously.  C{None} for no limit.

    @ivar backlog: The maximum number of C{Deferred} gets to allow at one
    time.  When an attempt is made to get an object which would exceed this
    limit, C{defer.QueueUnderflow} is raised synchronously.  C{None} for no
    limit.
    """

    def __init__(self, size=None, backlog=None):
        self.size = size
        self.backlog = backlog
        self.waiting = deque()  # Deferreds awaiting an object.
        self._counter = itertools.count(1)  # Unique sequence count.
        self._objFinder = {}  # Mapping of queued objects to tasks.
        self.clear()

    def clear(self):
        """
        Clear the pending queue.
        """
        # self._pending is a heapq of queued C{Task} instances
        # (some of which may be marked as invalid).
        self._pending = []
        self._nInvalid = 0  # The number of invalid (deleted) queued instances.

    def put(self, obj, priority, count=None):
        """
        Add an object to this queue.

        @param priority: the priority for the object. Lower is more
        important.

        @param obj: the object to add to the queue.

        @param count: A count used to maintain a stable (sort) order of
        tasks that have the same priority.

        @raise defer.QueueOverflow: Too many objects are in the queue.

        @raise RuntimeError: If this object already exists in the queue.
        """
        if self.waiting:
            self.waiting.popleft().callback(obj)
        elif self.size is None or len(self) < self.size:
            try:
                existing = self._objFinder[obj]
            except KeyError:
                pass
            else:
                if existing.valid:
                    raise RuntimeError("Object %r already in queue." % (obj,))
            if count is None:
                count = next(self._counter)
            task = _Task(obj, priority, count)
            self._objFinder[obj] = task
            heappush(self._pending, task)
        else:
            raise defer.QueueOverflow()

    def get(self):
        """
        Attempt to retrieve and remove the highest priority object from the
        queue.

        @return: a L{Deferred} which fires with the highest priority object
        available in the queue.

        @raise QueueUnderflow: Too many (more than C{backlog})
        L{Deferred}s are already waiting for an object from this queue.
        """
        while self._pending:
            task = heappop(self._pending)
            del self._objFinder[task.obj]
            if not task.valid:
                self._nInvalid -= 1
            else:
                return defer.succeed(task.obj)
        if self.backlog is None or len(self.waiting) < self.backlog:
            d = defer.Deferred(canceller=self._cancelGet)
            self.waiting.append(d)
            return d
        else:
            raise defer.QueueUnderflow()

    def delete(self, obj):
        """
        Delete an object from the queue.

        @raise: C{KeyError} if the object is not in the queue or is not
        valid.

        @param obj: The object to delete.
        """
        task = self._objFinder[obj]
        if task.valid:
            self._nInvalid += 1
            task.valid = False
        else:
            raise KeyError(obj)

    def reprioritize(self, obj, priority):
        """
        Set the priority of a pending task to priority.

        @raise: C{KeyError} if the object is not in the queue or is not
        valid.

        @param obj: The object whose priority should be altered.

        @param priority: The new priority. Lower is more important.
        """
        existing = self._objFinder[obj]
        if existing.valid:
            if existing.priority != priority:
                count = existing.count
                # Note: set existing.valid to False before calling
                # self.put, or it will complain that the obj is already in
                # the queue.
                existing.valid = False
                self.put(obj, priority, count)
        else:
            raise KeyError(obj)

    def _cancelGet(self, d):
        """
        Remove a deferred d from our waiting list, as the deferred has been
        canceled.

        Note: We do not need to wrap this in a try/except to catch d not
        being in self.waiting because this canceller will not be called if
        d has fired. put() pops a deferred out of self.waiting and calls
        it, so the canceller will no longer be called.

        @param d: The deferred that has been canceled.
        """
        self.waiting.remove(d)

    def asGenerator(self):
        """
        Return the queue contents as a generator, excluding all invalid
        entries.

        Note: the queued objects are returned in heap order, which is not
        the same as sorted order. If you need the objects sorted by
        priority, you can call sort on the result.
        """
        return (task.obj for task in self._pending if task.valid)

    def __len__(self):
        """
        The number of things in the queue is the length of the pending
        list, leaving out those tasks marked as invalid.
        """
        return len(self._pending) - self._nInvalid



from twisted.trial import unittest
from twisted.internet import defer, reactor, task
from txrdq.rdq import ResizableDispatchQueue, QueueStopped
from txrdq.job import Job


class TestSimple(unittest.TestCase):
    """
    Simple tests of the ResizableDispatchQueue class.
    """
    def testUnstarted(self):
        dq = ResizableDispatchQueue(None)
        self.assertEqual(dq.pending(), [])
        self.assertEqual(dq.underway(), set())
        self.assertFalse(dq.stopped)
        self.assertFalse(dq.paused)
        self.assertEqual((0, 0), dq.size())
        return dq.stop()

    def testNumericStartWidth(self):
        dq = ResizableDispatchQueue(None, 10)
        self.assertEqual(dq.width, 10)

    def testNegativeStartWidth(self):
        self.assertRaises(ValueError, ResizableDispatchQueue, None, -1)

    def testNonNumericStartWidth(self):
        self.assertRaises(ValueError, ResizableDispatchQueue, None, 'hi')

    def testStartedSize(self):
        dq = ResizableDispatchQueue(None)
        map(dq.put, range(3))
        self.assertEqual((0, 3), dq.size())

    @defer.inlineCallbacks
    def testStopSetsStopped(self):
        dq = ResizableDispatchQueue(None, 5)
        yield dq.stop()
        self.assertEqual(True, dq.stopped)

    @defer.inlineCallbacks
    def testStopAfterNoPuts(self):
        dq = ResizableDispatchQueue(None, 5)
        pending = yield dq.stop()
        self.assertEqual(pending, [])

    def testClearQueue(self):
        dq = ResizableDispatchQueue(None)
        map(dq.put, range(3))
        self.assertEqual((0, 3), dq.size())
        dq.clearQueue(cancelPending=False)
        self.assertEqual((0, 0), dq.size())

    @defer.inlineCallbacks
    def testClearQueueCancelPending(self):
        """
        When the queue is cleared with cancelPending=True, the pending jobs
        should fail and receive a job whose state is CANCELLED.
        """
        def ok(result):
            self.fail('Unexpected success!')

        def checkCancel(failure):
            self.assertEqual(failure.value.state, Job.CANCELLED)

        dq = ResizableDispatchQueue(None, 2)
        dq.put(0).addCallbacks(ok, checkCancel)
        dq.put(1).addCallbacks(ok, checkCancel)
        self.assertEqual((0, 2), dq.size())
        dq.clearQueue(cancelPending=True)
        self.assertEqual((0, 0), dq.size())
        yield dq.stop()

    @defer.inlineCallbacks
    def testPutRaisesWhenStopped(self):
        """
        Putting something onto a stopped queue should raise
        C{QueueStopped}.
        """
        dq = ResizableDispatchQueue(None)
        yield dq.stop()
        self.assertRaises(QueueStopped, dq.put, None)

    def testWidenWithNegativeWidth(self):
        """
        Widening a queue with a negative width should raise C{ValueError}.
        """
        dq = ResizableDispatchQueue(None)
        self.assertRaises(ValueError, dq.setWidth, -3)

    def testWidenWithNonNumericWidth(self):
        """
        Widening a queue with a non-numeric width should raise
        C{ValueError}.
        """
        dq = ResizableDispatchQueue(None)
        self.assertRaises(ValueError, dq.setWidth, 'Wider, pussycat! Wider!')

    @defer.inlineCallbacks
    def testGetPausedWidth(self):
        dq = ResizableDispatchQueue(None, 1)
        yield dq.pause()
        self.assertEqual(dq.width, 1)
        dq.setWidth(3)
        self.assertEqual(dq.width, 3)

    def testNarrowWiden(self):
        """
        A tests that exercises the code to reduce the number of pending
        stops in the queue. Narrowing to 7 means there will be 3 pending
        stops, then widening to 8 will decrease the number of pending stops
        by one.
        """
        dq = ResizableDispatchQueue(None, 10)
        dq.setWidth(7)
        self.assertEqual(dq.pendingStops, 3)
        self.assertEqual(dq.width, 7)
        dq.setWidth(8)
        self.assertEqual(dq.pendingStops, 2)
        self.assertEqual(dq.width, 8)

    @defer.inlineCallbacks
    def testPauseResumeWithZeroWidth(self):
        """
        Check that it is possible to resume with zero width.
        """
        dq = ResizableDispatchQueue(None, 1)
        yield dq.pause()
        dq.resume(0)
        self.assertEqual(dq.width, 0)

    @defer.inlineCallbacks
    def testPauseResumeWithNonZeroWidth(self):
        """
        Check that it is possible to resume with a non-zero width.
        """
        dq = ResizableDispatchQueue(None, 1)
        yield dq.pause()
        dq.resume(10)
        self.assertEqual(dq.width, 10)

    @defer.inlineCallbacks
    def testPauseResumeWithNegativeWidth(self):
        dq = ResizableDispatchQueue(None)
        yield dq.pause()
        self.assertRaises(ValueError, dq.resume, -5)

    @defer.inlineCallbacks
    def testPauseResumeWithNonNumericWidth(self):
        dq = ResizableDispatchQueue(None)
        yield dq.pause()
        self.assertRaises(ValueError, dq.resume, 'go!')

    @defer.inlineCallbacks
    def testDeferredFromPutFiresWithCorrectResult(self):
        """
        Make sure the C{Deferred} we get back from a put to the queue fires
        with the expected result.
        """
        dq = ResizableDispatchQueue(lambda x: 2 * x, width=1)
        job = yield dq.put(6)
        self.assertEqual(job.result, 12)

    def testReprioritizeNonExistentJob(self):
        """
        Deleting a non-existent job must raise C{KeyError}.
        """
        dq = ResizableDispatchQueue(None, 1)
        self.assertRaises(KeyError, dq.reprioritize, Job('x', 1), 10)

    @defer.inlineCallbacks
    def testReprioritizeNonPendingJob(self):
        """
        Reprioritizing a job that is not pending must raise
        C{RuntimeError}.
        """
        dq = ResizableDispatchQueue(lambda x: x, 1)
        job = yield dq.put('a')
        self.assertRaises(RuntimeError, dq.reprioritize, job, 10)


class TestTiming(unittest.TestCase):
    """
    Tests of the ResizableDispatchQueue class that involve testing the
    timing of events.
    """
    def slow(self, jobarg):
        return task.deferLater(reactor, 0.2, lambda x: x, jobarg)

    @defer.inlineCallbacks
    def _stopAndTest(self, when, dq, expectedPending):
        pendingJobs = yield task.deferLater(reactor, when, dq.stop)
        pendingArgs = [p.jobarg for p in pendingJobs]
        self.assertEqual(sorted(expectedPending), sorted(pendingArgs))
        self.assertEqual(dq.size(), (0, len(pendingJobs)))

    def _testPending(self, dq, expected):
        pendingJobs = dq.pending()
        pendingArgs = [p.jobarg for p in pendingJobs]
        self.assertEqual(sorted(expected), sorted(pendingArgs))

    def _testSize(self, dq, expected):
        self.assertEqual(expected, dq.size())

    def _testPaused(self, dq):
        self.assertTrue(dq.paused)

    def _testNotPaused(self, dq):
        self.assertFalse(dq.paused)

    def _testUnderway(self, dq, expected):
        self.assertEqual(expected, dq.underway())

    def testJustOne(self):
        dq = ResizableDispatchQueue(self.slow, 1)
        map(dq.put, range(3))
        # This should finish in about 0.2 seconds, with 2 things still on
        # the queue because we stop it after 0.001 so task 0 is already
        # executing but 1 and 2 are queued.
        return self._stopAndTest(0.001, dq, [1, 2])

    def testJustOneWithZeroInitialWidth(self):
        # Same as testJustOne (above), except we initialize with zero width
        # and start the dispatcher by explicitly setting its width.
        dq = ResizableDispatchQueue(self.slow)
        map(dq.put, range(3))
        self.assertEqual(0, dq.width)
        dq.width = 1
        return self._stopAndTest(0.0001, dq, [1, 2])

    def testOneUnderway(self):
        dq = ResizableDispatchQueue(self.slow, 1)
        map(dq.put, range(3))
        reactor.callLater(0.01, self._testUnderway, dq, set([0]))
        return self._stopAndTest(0.02, dq, [1, 2])

    @defer.inlineCallbacks
    def testStopAndCancelWithOneUnderway(self):
        """
        Start a dispatch queue of width 2, and send it 3 jobs. Verify that
        2 of the jobs are underway. Then stop it before they can complete,
        telling it to cancel the underway jobs. The two jobs that were
        underway should both be cancelled and returned by the stop method.
        The first 2 jobs returned should have state CANCELLED, and the
        final one should still be PENDING.
        """
        def ok(result):
            self.fail('Unexpected success!')

        def checkCancel(failure):
            self.assertEqual(failure.value.state, Job.CANCELLED)

        dq = ResizableDispatchQueue(self.slow, 2)
        dq.put(0).addCallbacks(ok, checkCancel)
        dq.put(1).addCallbacks(ok, checkCancel)
        dq.put(2)
        reactor.callLater(0.01, self._testUnderway, dq, set([0, 1]))
        pendingJobs = yield task.deferLater(
            reactor, 0.1, dq.stop, cancelUnderway=True)
        pendingArgs = [p.jobarg for p in pendingJobs]
        self.assertEqual([0, 1, 2], sorted(pendingArgs))
        self.assertEqual(pendingJobs[0].state, Job.CANCELLED)
        self.assertEqual(pendingJobs[1].state, Job.CANCELLED)
        self.assertEqual(pendingJobs[2].state, Job.PENDING)

    def testSequential(self):
        dq = ResizableDispatchQueue(self.slow, 1)
        map(dq.put, range(3))
        # This should finish in about 0.6 seconds, with nothing still on
        # the queue because we stop after 0.5 seconds so all three tasks
        # will have been dispatched one after another.
        return self._stopAndTest(0.5, dq, [])

    def testParallel(self):
        dq = ResizableDispatchQueue(self.slow, 3)
        map(dq.put, range(3))
        # This should finish in about 0.2 seconds, with nothing on the
        # queue.  We stop after just 0.001 seconds. All three tasks should
        # be dispatched immediately.
        return self._stopAndTest(0.001, dq, [])

    def testWiden(self):
        dq = ResizableDispatchQueue(self.slow, 2)
        map(dq.put, range(8))
        reactor.callLater(0.1, dq.setWidth, 10)
        # This should finish in about 0.4 seconds, with nothing on the
        # queue. We stop after 0.25 seconds. All tasks will have been
        # dispatched (width is initially 2 but is increased to 10 after the
        # first pair of jobs are underway).
        return self._stopAndTest(0.25, dq, [])

    def testNarrow(self):
        dq = ResizableDispatchQueue(self.slow, 2)
        map(dq.put, range(8))
        reactor.callLater(0.1, dq.setWidth, 1)
        # This should finish in about 0.4 seconds, with 3-7 still on the
        # queue. 0 and 1 are dispatched immediately, then we narrow to
        # width one, so only 2 is then dispatched before the stop happens.
        return self._stopAndTest(0.3, dq, [3, 4, 5, 6, 7])

    def testNarrowNarrow(self):
        dq = ResizableDispatchQueue(self.slow, width=3)
        map(dq.put, range(8))
        reactor.callLater(0.1, dq.setWidth, 2)
        reactor.callLater(0.3, dq.setWidth, 1)
        # This should finish in about 0.6 seconds, with 6 & 7 still on the
        # queue. 0, 1 and 2 are dispatched immediately, then we narrow to
        # width two, so 3 and 4 are dispatched in the next phase. Then we
        # narrow to one and 5 is dispatched before we stop at 0.5.
        return self._stopAndTest(0.5, dq, [6, 7])

    def testNarrowNarrowWiden(self):
        dq = ResizableDispatchQueue(self.slow, 3)
        map(dq.put, range(11))
        reactor.callLater(0.1, dq.setWidth, 2)
        reactor.callLater(0.3, dq.setWidth, 1)
        reactor.callLater(0.7, dq.setWidth, 3)
        return self._stopAndTest(0.9, dq, [10])

    def testNarrowNarrowWidenNarrow(self):
        items = range(14)
        dq = ResizableDispatchQueue(self.slow, 3)
        map(dq.put, items)
        # jobs 0, 1, 2 will be dispatched at time 0.00
        reactor.callLater(0.1, dq.setWidth, 2)
        reactor.callLater(
            0.1, self._testPending, dq, [3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13])
        # At time 0.20 jobs 0, 1, and 2 will all finish, but the
        # queue now has width 2, so 3 and 4 will be dispatched.
        reactor.callLater(
            0.25, self._testPending, dq, [5, 6, 7, 8, 9, 10, 11, 12, 13])
        reactor.callLater(0.3, dq.setWidth, 1)
        # 3 and 4 will finish at 0.40, and the queue is now width 1, so 5
        # will be dispatched at 0.40. Then 6 will be dispatched at 0.60.
        reactor.callLater(0.7, dq.setWidth, 3)
        # At 0.7 the queue is widened to 3, so 2 more jobs (7, 8) will be
        # dispatched. 6 will finish at 0.80, so one more job (9) will be
        # dispatched at 0.8.
        reactor.callLater(0.85, dq.setWidth, 1)
        # 7 and 8 will have finished at 0.9, at which point the queue is
        # down to size 1, but 9 is still in progress. At 1.0, 9 will finish
        # and 10 will launch. At 1.20, 10 will finish and 11 will be
        # dispatched.  11 will finish at about 1.40.
        return self._stopAndTest(1.35, dq, [12, 13])

    @defer.inlineCallbacks
    def testPause(self):
        items = range(3)
        dq = ResizableDispatchQueue(self.slow, 1)
        map(dq.put, items)
        reactor.callLater(0.2, self._testPaused, dq)
        yield task.deferLater(reactor, 0.1, dq.pause)
        pendingJobs = dq.pending()
        self.assertEqual([p.jobarg for p in pendingJobs], [1, 2])

    @defer.inlineCallbacks
    def testPauseStop(self):
        items = range(3)
        dq = ResizableDispatchQueue(self.slow, 1)
        map(dq.put, items)
        d1 = task.deferLater(reactor, 0.1, dq.pause)
        d2 = self._stopAndTest(0.15, dq, [1, 2])
        yield d1
        yield d2

    @defer.inlineCallbacks
    def testPauseAddStop(self):
        items = range(3)
        dq = ResizableDispatchQueue(self.slow, 1)
        map(dq.put, items)
        d1 = task.deferLater(reactor, 0.1, dq.pause)
        reactor.callLater(0.12, dq.put, None)
        reactor.callLater(0.14, dq.put, 'hello')
        reactor.callLater(0.16, dq.put, 'world')
        d2 = self._stopAndTest(0.18, dq, [1, 2, None, 'hello', 'world'])
        yield d1
        yield d2

    @defer.inlineCallbacks
    def testPauseResumeStop(self):
        """
        Pause a queue, then resume it, then stop it and check its contents.
        """
        items = range(3)
        dq = ResizableDispatchQueue(self.slow, 1)
        map(dq.put, items)
        d1 = task.deferLater(reactor, 0.1, dq.pause)
        reactor.callLater(0.5, dq.resume)
        reactor.callLater(0.51, self._testNotPaused, dq)
        d2 = self._stopAndTest(0.6, dq, [2])
        yield d1
        yield d2

    @defer.inlineCallbacks
    def testPauseStopResume(self):
        """
        Pause a queue, then stop it, then check that trying to resume it
        gets a C{QueueStopped} exception.
        """
        items = range(3)
        dq = ResizableDispatchQueue(self.slow, 1)
        map(dq.put, items)
        d1 = task.deferLater(reactor, 0.1, dq.pause)
        d2 = self._stopAndTest(0.2, dq, [1, 2])
        d3 = task.deferLater(reactor, 0.3, dq.resume)
        yield d1
        yield d2
        yield self.failUnlessFailure(d3, QueueStopped)

    @defer.inlineCallbacks
    def testPauseWidenResumeStop(self):
        items = range(3)
        dq = ResizableDispatchQueue(self.slow, 1)
        map(dq.put, items)
        d1 = task.deferLater(reactor, 0.1, dq.pause)
        reactor.callLater(0.2, dq.setWidth, 2)
        reactor.callLater(0.5, dq.resume)
        d2 = self._stopAndTest(0.6, dq, [])
        yield d1
        yield d2

    @defer.inlineCallbacks
    def testPausePlayAroundResumeStop(self):
        # As above, with some playing around while paused.
        items = range(3)
        dq = ResizableDispatchQueue(self.slow, 1)
        map(dq.put, items)
        d1 = task.deferLater(reactor, 0.1, dq.pause)
        reactor.callLater(0.12, dq.setWidth, 100)
        reactor.callLater(0.16, dq.setWidth, 0)
        reactor.callLater(0.2, dq.setWidth, 2)
        reactor.callLater(0.5, dq.resume)
        d2 = self._stopAndTest(0.6, dq, [])
        yield d1
        yield d2

    def testSizeAfterStarting(self):
        dq = ResizableDispatchQueue(self.slow, 1)
        map(dq.put, range(3))
        reactor.callLater(0.04, self._testSize, dq, (1, 2))
        return self._stopAndTest(0.1, dq, [1, 2])

    def testSizeAfterWiden(self):
        """
        Do some queue size testing while adjusting queue width.
        """
        dq = ResizableDispatchQueue(self.slow, 2)
        map(dq.put, range(8))
        reactor.callLater(0.02, self._testSize, dq, (2, 6))
        reactor.callLater(0.05, dq.setWidth, 4)
        reactor.callLater(0.07, self._testSize, dq, (4, 4))
        reactor.callLater(0.1, dq.setWidth, 10)
        reactor.callLater(0.15, self._testSize, dq, (8, 0))
        reactor.callLater(0.24, self._testSize, dq, (6, 0))
        reactor.callLater(0.28, self._testSize, dq, (4, 0))
        reactor.callLater(0.33, self._testSize, dq, (0, 0))
        # We stop after 0.35 seconds. All tasks will have been dispatched
        # (width is initially 2 but is increased to 10 after the first pair
        # of jobs are underway).
        return self._stopAndTest(0.35, dq, [])

    @defer.inlineCallbacks
    def testCancelRemovesFromQueue(self):
        """
        Cancelling a job should remove it from the queue.
        """
        def eb(failure):
            self.assertEqual(failure.value.state, Job.CANCELLED)

        dq = ResizableDispatchQueue(self.slow, 1)
        dq.put('a', 1)
        deferred = dq.put('b', 2)
        deferred.addErrback(eb)
        deferred.cancel()
        yield task.deferLater(reactor, 0.1, dq.pause)
        pendingJobs = dq.pending()
        self.assertEqual(pendingJobs, [])



"""
The ResizableDispatchQueue ...
"""

import time

from twisted.internet import task
from twisted.python import log

from txrdq.dpq import DeferredPriorityQueue
from txrdq.pool import DeferredPool
from txrdq.job import Job


class QueueStopped(Exception):
    """
    An exception that is raised when a client does a 'put' on a
    L{ResizableDispatchQueue} that has been permanently stopped.
    """


class ResizableDispatchQueue(object):
    """
    @ivar func: The one-argument function that will be called to process
    each job added to the queue.

    @ivar width: The 'width' of the queue. This is the maximum number of
    jobs that the queue will have in progress at any time. Set to 0
    to pause queue processing.

    @ivar size: The maximum number of objects to allow into the queue at a
    time. See the documentation for twisted.internet.defer.DeferredQueue

    @ivar backlog: The maximum number of L{Deferred} gets to allow at one
    time.  See the documentation for twisted.internet.defer.DeferredQueue.

    @raise ValueError: if width is not an integer or is less than zero.
    """

    NO_JOB = object()

    def __init__(self, func, width=0, size=None, backlog=None):
        self._func = func
        self.stopped = self.paused = False
        self._queue = DeferredPriorityQueue(size, backlog)
        self._pool = DeferredPool()
        self._coop = task.Cooperator()
        self._currentWidth = 0
        self.pendingStops = 0
        self._underway = set()
        self.width = width
        self.init = time.time()

    def put(self, jobarg, priority=0):
        """
        Add a job to the queue.

        @param jobarg: The argument that will be passed to self._func when
        this job is launched.

        @param priority: The priority for the job. Lower means more
        important.  Default is zero.

        @raise QueueStopped: if the queue has been stopped.

        @return: a C{Deferred} that will eventually fire with the result of
        calling self._func on jobarg.
        """
        if self.stopped:
            raise QueueStopped()
        else:
            job = Job(jobarg, priority)
            self._queue.put(job, priority)
            return job.watch().addBoth(self._jobDone, job)

    def _jobDone(self, result, job):
        """
        A job has finished one way or another. 'result' is either a
        successful (C{Job}) result or a failure (containing a C{Job} that
        failed or was cancelled). The job argument contains the original
        job in any case.

        Remove a finished job from our queue or (if it's no longer there)
        from the set of underway jobs.

        @param result: the result of the job processing.

        @param job: the job that has just finished.

        @return: the result we are passed.
        """
        try:
            self._queue.delete(job)
        except KeyError:
            self._underway.discard(job)
        return result

    def pending(self):
        """
        Get the pending jobs.

        @return: A C{list} of the L{Job}s that are pending (i.e., which
        have not yet begun to be processed).
        """
        # Turn the generator from our pending queue into a list, to make
        # sure we're returning a snapshot of the queue at this moment.
        return list(self._queue.asGenerator())

    def underway(self):
        """
        Get the currently underway jobs.

        @return: The set of jobs that are underway.
        """
        return set(job.jobarg for job in self._underway)

    def clearQueue(self, cancelPending=True):
        """
        Clear the pending list of jobs.

        @param cancelPending: if True, call 'cancel' on the C{Deferred}s
        for the pending jobs.
        """
        if cancelPending:
            jobs = list(self._queue.asGenerator())
            for job in jobs:
                job.cancel()
        self._queue.clear()

    def size(self):
        """
        Provide information about our size.

        @return: A two-tuple with the number of jobs that are underway and
        the number that are pending.
        """
        return (len(self._underway), len(self._queue))

    def _drain(self):
        """
        For a queue that is either stopped or paused, return a deferred
        that fires when the currently underway jobs in the queue have all
        finished.

        @raises RuntimeError: if the queue is not paused or stopped.

        @return: a C{Deferred} that fires when no jobs are underway.
        """
        # The queue must be paused or stopped in order for us to sanely
        # wait for it to drain to zero.
        if not (self.paused or self.stopped):
            raise RuntimeError('Queue is not paused or stopped.')

        # Flush idle job dispatcher (Deferreds) by giving each a no-op job.
        while self._queue.waiting:
            # We pass an arbitrary priority of zero. We know the queue is
            # empty because self._queue.waiting is not empty, so task
            # priority is irrelevant.
            self._queue.put(self.NO_JOB, 0)

        return self._pool.notifyWhenEmpty()

    def stop(self, cancelUnderway=False):
        """
        Permanently stop the queue. Any subsequent attempt to put a job
        onto the queue will result in an error. The default behavior is to
        wait for underway jobs to finish, but see the docs below on
        cancelUnderway.

        @param cancelUnderway: if True: 1) call cancel on the deferreds for
        all the currently underway jobs, and 2) add those underway jobs to
        the list of pending jobs returned by the C{Deferred} we return.

        @return: a C{Deferred} that fires with the list of pending jobs.
        """
        self.stopped = True
        self.width = 0

        log.msg('Stop called. Underway in queue %r' % self._underway)

        if cancelUnderway:
            underwayJobs = list(self._underway)
            for job in underwayJobs:
                job.cancel()
        else:
            underwayJobs = []
        log.msg('stop called. underwayJobs is %r' % underwayJobs)
        d = self._drain()
        d.addCallback(lambda _: underwayJobs + self.pending())
        return d

    def reprioritize(self, job, priority):
        """
        Set the priority of a pending job to priority.

        @raise: C{RuntimeError} if the job is not PENDING.

        @raise: C{KeyError} if the job does not exist.

        @param job: The job whose priority should be altered.

        @param priority: The new priority. Lower is more important.
        """
        if job.state != Job.PENDING:
            raise RuntimeError('Cannot delete job that is not pending.')
        else:
            self._queue.reprioritize(job, priority)

    def pause(self):
        self._pausedWidth = self.width
        self.width = 0
        self.paused = True
        return self._drain()

    def resume(self, width=None):
        if self.stopped:
            raise QueueStopped()
        if width is None:
            width = self._pausedWidth
        else:
            width = int(width)
            if width < 0:
                raise ValueError('Queue width cannot be negative.')
        self.paused = False
        self.width = width

    def next(self):
        """
        Return the next job to be processed, unless we are currenly
        narrowing the queue, in which case we raise C{StopIteration}.

        @raise StopIteration: if the queue is in the process of being
        narrowed.

        @return: a C{Deferred} that will fire when a job is added to the
        queue.
        """
        if self.pendingStops:
            self.pendingStops -= 1
            self._currentWidth -= 1
            raise StopIteration
        else:
            return self._queue.get().addCallback(self._launchJob)

    def _launchJob(self, job):
        """
        Launch a queued job.

        @param job: a L{Job} to get underway.  If the job is a no-op, do
        nothing.

        @return: Unless the job is a no-op, return a C{Deferred} that will fire
        with the result of processing the job. Otherwise, return C{None}.
        """
        if job is not self.NO_JOB:
            self._underway.add(job)
            job.launch(self._func)
            # Return the first waiting Deferred for the job, if any. Our
            # self._coop (a task.Cooperator) will wait for it. If there is
            # no waiting Deferred, the job has already finished (probably
            # synchronously) and there's nothing to wait for.
            try:
                return job.waiting[0]
            except IndexError:
                pass

    def getWidth(self):
        """
        Return the current width of the queue.

        Note that this may be greater or less than the number of jobs
        currently underway. For example, the width could be 10 but we may
        have only 7 jobs underway (due to their being no more pending jobs
        to launch). If the width is then reduced to 4, this method will
        report width as 4 even though there are still 7 jobs in progress.

        @return: the current width of the queue.
        """
        if self.paused:
            return self._pausedWidth
        else:
            return self._currentWidth - self.pendingStops

    def setWidth(self, width):
        """
        Set the width of the queue. If the queue is paused, we simply
        remember what the width should be once queue processing is
        resumed. If we're not paused, then the interesting cases are either
        making the queue wider or narrower. To make ourselves wider, we
        pass 'self' to our coiterator the appropriate number of times. To
        make ourselves narrower, we arrange for self.next to raise
        StopIteration the appropriate number of times.

        @param width: the new width (a non-negative integer) for the queue.

        @raises ValueError: if width is not an integer or is less than zero.
        """
        width = int(width)
        if width < 0:
            raise ValueError('Queue width cannot be negative.')
        if self.paused:
            self._pausedWidth = width
        else:
            targetWidth = self._currentWidth - self.pendingStops
            extra = width - targetWidth
            if extra > 0:
                # Make ourselves wider.
                delta = extra - self.pendingStops
                if delta >= 0:
                    self.pendingStops = 0
                    for i in xrange(delta):
                        self._pool.add(self._coop.coiterate(self))
                    self._currentWidth += delta
                else:
                    self.pendingStops -= extra
            elif extra < 0:
                # Make ourselves narrower.
                self.pendingStops -= extra

    width = property(getWidth, setWidth)



#!/usr/bin/env python

import os
from distutils.core import setup


# Utility function to read the README file.
# Credit: http://pypi.python.org/pypi/an_example_pypi_project
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(name='txRDQ',
      version='0.2.4',
      provides=['txrdq'],
      requires=['twisted'],
      maintainer='Fluidinfo Inc.',
      maintainer_email='info@fluidinfo.com',
      url='https://launchpad.net/txrdq',
      download_url='https://code.launchpad.net/txrdq',
      packages=['txrdq', 'txrdq.test'],
      keywords=['twisted dispatch job queue'],
      classifiers=[
          "Programming Language :: Python",
          "Framework :: Twisted",
          "Development Status :: 4 - Beta",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: Apache Software License",
          "Operating System :: OS Independent",
          "Topic :: Software Development :: Libraries :: Python Modules",
          ],
      description=("A Twisted class for queueing and dispatching jobs in "
                   "a controlled manner."),
      long_description=read('README'),
      )
